from asyncio import Future
import asyncio
import logging
import random
from collections import defaultdict

from again.utils import unique_hex
from .packet import ControlPacket


class RegistryClient:
    logger = logging.getLogger(__name__)

    def __init__(self, loop, host, port, bus):
        self._bus = bus
        self._loop = loop
        self._host = host
        self._port = port
        self._transport = None
        self._protocol = None
        self._service = None
        self._version = None
        self._node_id = None
        self._pending_requests = {}
        self._available_services = defaultdict(list)
        self._assigned_services = defaultdict(lambda: defaultdict(list))

    def _register_service(self, ip, port, service, vendors, version, service_type):
        self._service = service
        self._version = version
        self._node_id = '{}_{}_{}'.format(service, version, unique_hex())
        packet = ControlPacket.registration(ip, port, self._node_id, service, version, vendors, service_type)
        self._protocol.send(packet)

    def get_instances(self, service, version):
        future = asyncio.Future()
        packet = ControlPacket.instances(service, version)
        self._protocol.send(packet)
        self._pending_requests[packet['request_id']] = future
        return future

    def register_http(self, vendors, ip, port, service, version):
        self._register_service(ip, port, service, vendors, version, 'http')

    def register_tcp(self, vendors, ip, port, service, version):
        self._register_service(ip, port, service, vendors, version, 'tcp')

    def subscribe_for_message(self, packet):
        packet['node_id'] = self._node_id
        self._protocol.send(packet)

    def _protocol_factory(self):
        from vyked.jsonprotocol import RegistryClientProtocol

        p = RegistryClientProtocol(self)
        return p

    def connect(self):
        coro = self._loop.create_connection(self._protocol_factory, self._host, self._port)
        self._transport, self._protocol = self._loop.run_until_complete(coro)

    def receive(self, packet: dict, registry_protocol):
        if packet['type'] == 'registered':
            self.cache_vendors(packet['params']['vendors'])
            self._bus.registration_complete()
        elif packet['type'] == 'deregister':
            self._handle_deregistration(packet)
        elif packet['type'] == 'subscription_list':
            self._handle_subscription_list(packet)
        elif packet['type'] == 'message_subscription_list':
            self._handle_subscription_list(packet)
        elif packet['type'] == 'instances':
            future = self._pending_requests.pop(packet['request_id'])
            future.set_result(packet['params']['instances'])

    def get_all_addresses(self, full_service_name):
        return self._available_services.get(
            self._get_full_service_name(full_service_name[0], full_service_name[1]))

    def get_for_node(self, node_id):
        for services in self._available_services.values():
            for host, port, node, service_type in services:
                if node == node_id:
                    return host, port, node, service_type
        return None

    def get_random_service(self, service_name, service_type):
        services = self._available_services[service_name]
        services = [service for service in services if service[3] == service_type]
        if len(services):
            return random.choice(services)
        else:
            return None

    def resolve(self, service: str, version: str, entity: str, service_type: str):
        service_name = self._get_full_service_name(service, version)
        if entity is not None:
            entity_map = self._assigned_services.get(service_name)
            if entity_map is None:
                self._assigned_services[service_name] = {}
            entity_map = self._assigned_services.get(service_name)
            if entity in entity_map:
                return entity_map[entity]
            else:
                host, port, node_id, service_type = self.get_random_service(service_name, service_type)
                if node_id is not None:
                    entity_map[entity] = host, port, node_id, service_type
                return host, port, node_id, service_type
        else:
            return self.get_random_service(service_name, service_type)

    @staticmethod
    def _get_full_service_name(service, version):
        return "{}/{}".format(service, version)

    def cache_vendors(self, vendors):
        for vendor in vendors:
            vendor_name = vendor['name']
            for address in vendor['addresses']:
                self._available_services[vendor_name].append(
                    (address['host'], address['port'], address['node_id'], address['type']))

    def _handle_subscription_list(self, packet):
        future = self._pending_requests.pop(packet['request_id'])
        future.set_result(packet['nodes'])

    def _handle_deregistration(self, packet):
        params = packet['params']
        vendor = params['vendor']
        node = params['node_id']
        for each in self._available_services[vendor]:
            if each[2] == node:
                self._available_services[vendor].remove(each)
        entity_map = self._assigned_services.get(vendor)
        if entity_map is not None:
            stale_entities = []
            for entity, node_id in entity_map.items():
                if node == node_id:
                    stale_entities.append(entity)
            for entity in stale_entities:
                entity_map.pop(entity)



import json
import logging
import asyncio

import asyncio_redis as redis
from vyked.utils.log import log

class PubSubHandler:
    _logger = logging.getLogger(__name__)

    def __init__(self, redis_host, redis_port):
        self._redis_host = redis_host
        self._redis_port = redis_port
        self._conn = None

    @asyncio.coroutine
    def connect(self):
        self._conn = yield from self._get_conn()
        return self._conn

    @asyncio.coroutine
    @log(logger=_logger)
    def publish(self, service, version, endpoint, payload):
        if self._conn is not None:
            try:
                yield from self._conn.publish(self._get_key(service, version, endpoint), json.dumps(payload))
                return True
            except redis.Error as e:
                self._logger.error('Publish failed with error %s', repr(e))
        return False

    @asyncio.coroutine
    def subscribe(self, endpoints, handler):
        connection = yield from self._get_conn()
        subscriber = yield from connection.start_subscribe()
        channels = [self._get_key(*endpoint) for endpoint in endpoints]
        yield from subscriber.subscribe(channels)
        while True:
            payload = yield from subscriber.next_published()
            service, version, endpoint = payload.channel.split('.')
            handler(service, int(version), endpoint, payload.value)
        return False

    @staticmethod
    def _get_key(service, version, endpoint):
        return '{}.{}.{}'.format(service, version, endpoint)

    def _get_conn(self):
        return (yield from redis.Connection.create(self._redis_host, self._redis_port, auto_reconnect=True))









from asyncio import coroutine
from functools import wraps
from enum import Enum

from aiopg import create_pool, Pool, Cursor
import psycopg2

_CursorType = Enum('CursorType', 'PLAIN, DICT, NAMEDTUPLE')


def dict_cursor(func):
    """
    Decorator that provides a dictionary cursor to the calling function

    Adds the cursor as the second argument to the calling functions

    Requires that the function being decorated is an instance of a class or object
    that yields a cursor from a get_cursor(cursor_type=CursorType.DICT) coroutine or provides such an object
    as the first argument in its signature

    Yields:
        A client-side dictionary cursor
    """

    @wraps(func)
    def wrapper(cls, *args, **kwargs):
        with (yield from cls.get_cursor(_CursorType.DICT)) as c:
            return (yield from func(cls, c, *args, **kwargs))

    return wrapper


def cursor(func):
    """
    Decorator that provides a cursor to the calling function

    Adds the cursor as the second argument to the calling functions

    Requires that the function being decorated is an instance of a class or object
    that yields a cursor from a get_cursor() coroutine or provides such an object
    as the first argument in its signature

    Yields:
        A client-side cursor
    """

    @wraps(func)
    def wrapper(cls, *args, **kwargs):
        with (yield from cls.get_cursor()) as c:
            return (yield from func(cls, c, *args, **kwargs))

    return wrapper


def nt_cursor(func):
    """
    Decorator that provides a namedtuple cursor to the calling function

    Adds the cursor as the second argument to the calling functions

    Requires that the function being decorated is an instance of a class or object
    that yields a cursor from a get_cursor(cursor_type=CursorType.NAMEDTUPLE) coroutine or provides such an object
    as the first argument in its signature

    Yields:
        A client-side namedtuple cursor
    """

    @wraps(func)
    def wrapper(cls, *args, **kwargs):
        with (yield from cls.get_cursor(_CursorType.NAMEDTUPLE)) as c:
            return (yield from func(cls, c, *args, **kwargs))

    return wrapper


def transaction(func):
    """
    Provides a transacted cursor which will run in autocommit=false mode

    For any exception the transaction will be rolled back.
    Requires that the function being decorated is an instance of a class or object
    that yields a cursor from a get_cursor(cursor_type=CursorType.NAMEDTUPLE) coroutine or provides such an object
    as the first argument in its signature

    Yields:
        A client-side transacted named cursor
    """

    @wraps(func)
    def wrapper(cls, *args, **kwargs):
        with (yield from cls.get_cursor(_CursorType.NAMEDTUPLE)) as c:
            try:
                yield from c.execute('BEGIN')
                result = (yield from func(cls, c, *args, **kwargs))
            except Exception:
                yield from c.execute('ROLLBACK')
            else:
                yield from c.execute('COMMIT')
                return result

    return wrapper


class PostgresStore:
    _pool = None
    _connection_params = {}
    _insert_string = "insert into {} ({}) values ({}) returning *;"
    _update_string = "update {} set ({}) = ({}) where ({}) returning *;"
    _select_all_string_with_condition = "select * from {} where ({}) order by {} limit {} offset {};"
    _select_all_string = "select * from {} order by {} limit {} offset {};"
    _select_selective_column = "select {} from {} order by {} limit {} offset {};"
    _select_selective_column_with_condition = "select {} from {} where ({}) order by {} limit {} offset {};"
    _delete_query = "delete from {} where ({});"
    _count_query = "select count(*) from {};"
    _count_query_where = "select count(*) from {} where {};"
    _OR = ' or '
    _AND = ' and '
    _LPAREN = '('
    _RPAREN = ')'
    _WHERE_AND = '{} {} %s'
    _PLACEHOLDER = ' %s,'
    _COMMA = ', '

    @classmethod
    def connect(cls, database:str, user:str, password:str, host:str, port:int):
        """
        Sets connection parameters
        """
        cls._connection_params['database'] = database
        cls._connection_params['user'] = user
        cls._connection_params['password'] = password
        cls._connection_params['host'] = host
        cls._connection_params['port'] = port

    @classmethod
    def use_pool(cls, pool:Pool):
        """
        Sets an existing connection pool instead of using connect() to make one
        """
        cls._pool = pool

    @classmethod
    @coroutine
    def get_pool(cls) -> Pool:
        """
        Yields:
            existing db connection pool
        """
        if len(cls._connection_params) < 5:
            raise ConnectionError('Please call SQLStore.connect before calling this method')
        if not cls._pool:
            cls._pool = yield from create_pool(minsize=1, maxsize=50, keepalives_idle=5, keepalives_interval=4, **cls._connection_params)
        return cls._pool

    @classmethod
    @coroutine
    def get_cursor(cls, cursor_type=_CursorType.PLAIN) -> Cursor:
        """
        Yields:
            new client-side cursor from existing db connection pool
        """
        pool = yield from cls.get_pool()
        if cursor_type == _CursorType.PLAIN:
            c = yield from pool.cursor()
            return c
        if cursor_type == _CursorType.NAMEDTUPLE:
            return (yield from pool.cursor(cursor_factory=psycopg2.extras.NamedTupleCursor))
        if cursor_type == _CursorType.DICT:
            return (yield from pool.cursor(cursor_factory=psycopg2.extras.DictCursor))

    @classmethod
    @coroutine
    @cursor
    def count(cls, cur, table:str, where_keys: list=None):
        """
        gives the number of records in the table

        Args:
            table: a string indicating the name of the table

        Returns:
            an integer indicating the number of records in the table

        """

        if where_keys:
            where_clause, values = cls._get_where_clause_with_values(where_keys)
            query = cls._count_query_where.format(table, where_clause)
            q, t = query, values
        else:
            query = cls._count_query.format(table)
            q, t = query, ()
        yield from cur.execute(q, t)
        result = yield from cur.fetchone()
        return int(result[0])

    @classmethod
    @coroutine
    @nt_cursor
    def insert(cls, cur, table: str, values: dict):
        """
        Creates an insert statement with only chosen fields

        Args:
            table: a string indicating the name of the table
            values: a dict of fields and values to be inserted

        Returns:
            A 'Record' object with table columns as properties

        """
        keys = cls._COMMA.join(values.keys())
        value_place_holder = cls._PLACEHOLDER * len(values)
        query = cls._insert_string.format(table, keys, value_place_holder[:-1])
        yield from cur.execute(query, tuple(values.values()))
        return (yield from cur.fetchone())

    @classmethod
    @coroutine
    @nt_cursor
    def update(cls, cur, table: str, values: dict, where_keys: list) -> tuple:
        """
        Creates an update query with only chosen fields
        Supports only a single field where clause

        Args:
            table: a string indicating the name of the table
            values: a dict of fields and values to be inserted
            where_keys: list of dictionary
            example of where keys: [{'name':('>', 'cip'),'url':('=', 'cip.com'},{'type':{'<=', 'manufacturer'}}]
            where_clause will look like ((name>%s and url=%s) or (type <= %s))
            items within each dictionary get 'AND'-ed and dictionaries themselves get 'OR'-ed

        Returns:
            an integer indicating count of rows deleted

        """
        keys = cls._COMMA.join(values.keys())
        value_place_holder = cls._PLACEHOLDER * len(values)
        where_clause, where_values = cls._get_where_clause_with_values(where_keys)
        query = cls._update_string.format(table, keys, value_place_holder[:-1], where_clause)
        yield from cur.execute(query, (tuple(values.values()) + where_values))
        return (yield from cur.fetchall())

    @classmethod
    def _get_where_clause_with_values(cls, where_keys):
        values = []

        def make_and_query(ele: dict):
            and_query = cls._AND.join([cls._WHERE_AND.format(e[0], e[1][0]) for e in ele.items()])
            values.extend([val[1] for val in ele.values()])
            return cls._LPAREN + and_query + cls._RPAREN

        return cls._OR.join(map(make_and_query, where_keys)), tuple(values)

    @classmethod
    @coroutine
    @cursor
    def delete(cls, cur, table: str, where_keys: list):
        """
        Creates a delete query with where keys
        Supports multiple where clause with and or or both

        Args:
            table: a string indicating the name of the table
            where_keys: list of dictionary
            example of where keys: [{'name':('>', 'cip'),'url':('=', 'cip.com'},{'type':{'<=', 'manufacturer'}}]
            where_clause will look like ((name>%s and url=%s) or (type <= %s))
            items within each dictionary get 'AND'-ed and dictionaries themselves get 'OR'-ed

        Returns:
            an integer indicating count of rows deleted

        """
        where_clause, values = cls._get_where_clause_with_values(where_keys)
        query = cls._delete_query.format(table, where_clause)
        yield from cur.execute(query, values)
        return cur.rowcount

    @classmethod
    @coroutine
    @nt_cursor
    def select(cls, cur, table: str, order_by: str, columns: list=None, where_keys: list=None, limit=100,
               offset=0):
        """
        Creates a select query for selective columns with where keys
        Supports multiple where claus with and or or both

        Args:
            table: a string indicating the name of the table
            order_by: a string indicating column name to order the results on
            columns: list of columns to select from
            where_keys: list of dictionary
            limit: the limit on the number of results
            offset: offset on the results

            example of where keys: [{'name':('>', 'cip'),'url':('=', 'cip.com'},{'type':{'<=', 'manufacturer'}}]
            where_clause will look like ((name>%s and url=%s) or (type <= %s))
            items within each dictionary get 'AND'-ed and across dictionaries get 'OR'-ed

        Returns:
            A list of 'Record' object with table columns as properties

        """
        if columns:
            columns_string = cls._COMMA.join(columns)
            if where_keys:
                where_clause, values = cls._get_where_clause_with_values(where_keys)
                query = cls._select_selective_column_with_condition.format(columns_string, table, where_clause,
                                                                           order_by, limit, offset)
                q, t = query, values
            else:
                query = cls._select_selective_column.format(columns_string, table, order_by, limit, offset)
                q, t = query, ()
        else:
            if where_keys:
                where_clause, values = cls._get_where_clause_with_values(where_keys)
                query = cls._select_all_string_with_condition.format(table, where_clause, order_by, limit, offset)
                q, t = query, values
            else:
                query = cls._select_all_string.format(table, order_by, limit, offset)
                q, t = query, ()

        yield from cur.execute(q, t)
        return (yield from cur.fetchall())

    @classmethod
    @coroutine
    @nt_cursor
    def raw_sql(cls, cur, query: str, values: tuple):
        """
        Run a raw sql query

        Args:
            query : query string to execute
            values : tuple of values to be used with the query

        Returns:
            result of query as list of named tuple

        """
        yield from cur.execute(query, values)
        return (yield from cur.fetchall())



import collections


class OrderedClassMembers(type):
    @classmethod
    def __prepare__(self, name, bases):
        return collections.OrderedDict()

    def __new__(self, name, bases, classdict):
        classdict['__ordered__'] = [key for key in classdict.keys() if key not in ('__module__', '__qualname__')]
        return type.__new__(self, name, bases, classdict)





import logging
import asyncio

import aiohttp

DEFAULT_TIMEOUT = 15
DEFAULT_PING_INTERVAL= 10


class Pinger:
    _logger = logging.getLogger(__name__)

    def __init__(self, ping_handler, loop, timeout=DEFAULT_TIMEOUT):
        self._ping_handler = ping_handler
        self._loop = loop
        self._count = 0
        self._timeout = timeout
        self._interval = DEFAULT_PING_INTERVAL
        self._timer = None
        self._tcp = None
        self._protocol = None
        self._node = None
        self._host = None
        self._port = None

    def register_tcp_service(self, protocol, node):
        self._tcp = True
        self._protocol = protocol
        self._node = node

    def register_http_service(self, host, port, node):
        self._tcp = False
        self._host = host
        self._port = port
        self._node = node

    @asyncio.coroutine
    def start_ping(self):
        if self._tcp:
            self._timer = self._loop.call_later(self._interval, self._send_timed_ping)
        else:
            url = 'http://{}:{}/ping'.format(self._host, self._port)
            yield from asyncio.sleep(self._interval)
            self._logger.debug('Pinging node {} at url: {}'.format(self._node, url))
            try:
                resp = yield from asyncio.wait_for(aiohttp.request('GET', url=url), self._interval)
                resp.close()
                yield from self.start_ping()
            except:
                self._ping_handler.handle_ping_timeout(self._node)
                pass

    def pong_received(self, count):
        if self._timer is not None:
            self._timer.cancel()

        if self._count == count:
            self._count += 1
            yield from self.start_ping()
        else:
            self._ping_timed_out()

    def _make_ping_packet(self):
        packet = {'type': 'ping', 'node_id': self._node, 'count': self._count}
        return packet

    def _send_timed_ping(self):
        packet = self._make_ping_packet()
        self._protocol.send(packet)
        self._timer = self._loop.call_later(self._timeout, self._ping_timed_out)

    def _ping_timed_out(self):
        self._ping_handler.handle_ping_timeout(self._node)



import signal
import asyncio
from functools import partial
from collections import defaultdict

from again.utils import unique_hex

from .pinger import Pinger
from .utils.log import setup_logging


class Registry:
    def __init__(self, ip, port):
        self._ip = ip
        self._port = port
        self._loop = asyncio.get_event_loop()
        self._registered_services = defaultdict(list)
        self._pending_services = defaultdict(list)
        self._client_protocols = {}
        self._service_dependencies = {}
        self._service_protocols = {}
        self._dead_service_listeners = defaultdict(list)
        self._subscription_list = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        self._message_sub_list = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(list))))
        self._pingers = {}

    def _rfactory(self):
        from vyked.jsonprotocol import RegistryProtocol

        return RegistryProtocol(self)

    def start(self):
        setup_logging("registry")
        self._loop.add_signal_handler(getattr(signal, 'SIGINT'), partial(self._stop, 'SIGINT'))
        self._loop.add_signal_handler(getattr(signal, 'SIGTERM'), partial(self._stop, 'SIGTERM'))
        registry_coro = self._loop.create_server(self._rfactory, self._ip, self._port)
        self._server = self._loop.run_until_complete(registry_coro)
        try:
            self._loop.run_forever()
        except Exception as e:
            print(e)
        finally:
            self._server.close()
            self._loop.run_until_complete(self._server.wait_closed())
            self._loop.close()

    def _stop(self, signame:str):
        print('\ngot signal {} - exiting'.format(signame))
        self._loop.stop()

    def receive(self, packet:dict, registry_protocol, transport):
        request_type = packet['type']
        if request_type == 'register':
            self._register_service(packet, registry_protocol, *transport.get_extra_info('peername'))
        elif request_type == 'subscribe':
            self._handle_subscription(packet, *transport.get_extra_info('peername'))
        elif request_type == 'message_sub':
            self._subscribe_for_message(packet, *transport.get_extra_info('peername'))
        elif request_type == 'pong':
            self._handle_pong(packet['node_id'], packet['count'])
        elif request_type == 'resolve_publication':
            self._resolve_publication(packet, registry_protocol)
        elif request_type == 'resolve_message_publication':
            self._resolve_message_publication(packet, registry_protocol)
        elif request_type == 'service_death_sub':
            self._add_service_death_listener(packet)
        elif request_type == 'get_instances':
            self._get_service_instances(packet, registry_protocol)

    def deregister_service(self, node_id):
        for service, nodes in self._registered_services.items():
            service_present = False
            for node in nodes:
                if node[2] == node_id:
                    nodes.remove(node)
                    service_present = True
                    self._service_protocols.pop(node_id, None)
                    self._client_protocols.pop(node_id, None)
            if service_present:
                self._notify_consumers(service, node_id)
                self._remove_service_death_listener(node_id)
                self._notify_service_death_listeners(service, node_id)
                if not len(nodes):
                    for consumer in self._get_consumers(service):
                        self._pending_services[consumer] = [node_id for host, port, node_id, service_type in
                                                            self._registered_services[consumer]]


    def handle_ping_timeout(self, node_id):
        print("{} timed out".format(node_id))
        self.deregister_service(node_id)

    def _handle_pending_registrations(self):
        for service_name in self._pending_services:
            depends_on = self._service_dependencies[service_name]
            should_activate = True
            for dependency in depends_on:
                if self._registered_services.get(
                        self._get_full_service_name(dependency["service"], dependency["version"])) is None:
                    should_activate = False
                    break
            nodes = self._pending_services[service_name]
            for node in nodes:
                if should_activate:
                    self._send_activated_packet(self._client_protocols[node],
                                                self._service_dependencies[service_name])
                    nodes.remove(node)

    def _register_service(self, packet:dict, registry_protocol, host, port):
        params = packet['params']
        service_name = self._get_full_service_name(params["service"], params['version'])
        dependencies = params['vendors']
        service_entry = (host, params['port'], params['node_id'], params['type'])
        self._registered_services[service_name].append(service_entry)
        self._pending_services[service_name].append(params['node_id'])
        self._client_protocols[params['node_id']] = registry_protocol
        self._connect_to_service(host, params['port'], params['node_id'], params['type'])
        if self._service_dependencies.get(service_name) is None:
            self._service_dependencies[service_name] = dependencies
        self._handle_pending_registrations()

    @staticmethod
    def _get_full_service_name(service: str, version):
        return "{}/{}".format(service, version)

    def _send_activated_packet(self, protocol, dependencies):
        packet = self._make_activated_packet(dependencies)
        protocol.send(packet)
        pass

    def _make_activated_packet(self, vendors):
        vendors_packet = []
        for vendor in vendors:
            vendor_packet = defaultdict(list)
            vendor_name = self._get_full_service_name(vendor['service'], vendor['version'])
            for host, port, node, service_type in self._registered_services[vendor_name]:
                vendor_node_packet = {
                    'host': host,
                    'port': port,
                    'node_id': node,
                    'type': service_type
                }
                vendor_packet['name'] = vendor_name
                vendor_packet['addresses'].append(vendor_node_packet)
            vendors_packet.append(vendor_packet)
        params = {
            'vendors': vendors_packet
        }
        packet = {'pid': unique_hex(),
                  'type': 'registered',
                  'params': params}
        return packet

    def _connect_to_service(self, host, port, node_id, service_type):
        if service_type == 'tcp':
            coro = self._loop.create_connection(self._rfactory, host, port)
            future = asyncio.async(coro)
            future.add_done_callback(partial(self._handle_service_connection, node_id))
        else:
            pinger = Pinger(registry, self._loop)
            pinger.register_http_service(host, port, node_id)
            self._pingers[node_id] = pinger
            asyncio.async(pinger.start_ping())

    def _handle_service_connection(self, node_id, future):
        transport, protocol = future.result()
        self._service_protocols[node_id] = protocol
        pinger = Pinger(registry, self._loop)
        pinger.register_tcp_service(protocol, node_id)
        self._pingers[node_id] = pinger
        asyncio.async(pinger.start_ping())

    def _handle_subscription(self, packet, host, port):
        params = packet['params']
        for each in params['subscribe_to']:
            self._subscription_list[each['service']][each['version']][each['endpoint']].append(
                (host, params['port'], params['node_id']))

    def _handle_pong(self, node_id, count):
        pinger = self._pingers[node_id]
        asyncio.async(pinger.pong_received(count))

    def _resolve_publication(self, packet, protocol):
        params = packet['params']
        service, version, endpoint = params['service'], params['version'], params['endpoint']
        nodes = [{'ip': ip, 'port': port, 'node_id': node} for ip, port, node in
                 self._subscription_list[service][version][endpoint]]
        result = {'type': 'subscription_list', 'request_id': packet['request_id'], 'nodes': nodes}
        protocol.send(result)

    def _resolve_message_publication(self, packet, protocol):
        params = packet['params']
        service, version, endpoint, entity = params['service'], params['version'], params['endpoint'], params['entity']
        nodes = [{'ip': ip, 'port': port, 'node_id': node} for ip, port, node in
                 self._message_sub_list[service][version][endpoint][entity]]
        result = {'type': 'message_subscription_list', 'request_id': packet['request_id'], 'nodes': nodes}
        protocol.send(result)

    def _get_consumers(self, vendor):
        consumers = []
        for service, vendors in self._service_dependencies.items():
            for each in vendors:
                vendor_name = self._get_full_service_name(each['service'], each['version'])
                if vendor == vendor_name:
                    consumers.append(service)
        return consumers

    def _notify_consumers(self, vendor, node_id):
        packet = self._make_deregister_packet(node_id, vendor)
        for consumer in self._get_consumers(vendor):
            for host, port, node, service_type in self._registered_services[consumer]:
                protocol = self._client_protocols[node]
                protocol.send(packet)

    def _make_deregister_packet(self, node_id, vendor):
        packet = {'type': 'deregister'}
        params = {'node_id': node_id, 'vendor': vendor}
        packet['params'] = params
        return packet

    def _subscribe_for_message(self, packet, host, port):
        service, version, endpoint, entity = packet['service'], packet['version'], packet[
            'endpoint'], packet['entity']
        ip, port, node_id = packet['ip'], packet['port'], packet['node_id']
        self._message_sub_list[service][version][endpoint][entity].append((host, port, node_id))

    def _add_service_death_listener(self, packet):
        service, version, node = packet['service'], packet['version'], packet['node']
        params = packet['params']
        service_name = self._get_full_service_name(params['service'], params['version'])
        self._dead_service_listeners[service_name].append(node)

    def _remove_service_death_listener(self, node_id):
        for service, listeners in self._dead_service_listeners.items():
            if node_id in listeners:
                listeners.remove()

    def _notify_service_death_listeners(self, service, node_id):
        for node in self._dead_service_listeners[service]:
            protocol = self._client_protocols[node]
            protocol.send(self._make_service_dead_packet(service, node_id))

    @staticmethod
    def _make_service_dead_packet(service, node_id):
        params = {'node_id': node_id, 'service': service.split('/')[0], 'version': service.split('/')[1]}
        packet = {'type': 'service_dead', 'params': params}
        return packet

    def _get_service_instances(self, packet, registry_protocol):
        params = packet['params']
        service, version = params['service'], params['version']
        service_name = self._get_full_service_name(service, version)
        instances = []
        if service_name in self._registered_services:
            instances = [{'host': host, 'port': port, 'node': node, 'type': service_type} for host, port, node, service_type in
                         self._registered_services[service_name]]
        instance_packet_params = {'service': service, 'version': version, 'instances': instances}
        instance_packet = {'type': 'instances', 'request_id': packet['request_id'], 'params': instance_packet_params}
        registry_protocol.send(instance_packet)


if __name__ == '__main__':
    REGISTRY_HOST = None
    REGISTRY_PORT = 4500
    registry = Registry(REGISTRY_HOST, REGISTRY_PORT)
    registry.start()



class Entity:
    """
    objects which are defined by their identity
    """
    def __init__(self, id):
        self._id = id

    @property
    def id(self):
        return self._id


class Value:
    """
    describe a property of a thing, and they don’t have their own identity or lifecycle
    """
    @staticmethod
    def create(*args, **kwargs):
        pass


class Aggregate:
    """
    groups of objects which are treated as a single unit
    """
    pass


class Factory:
    """
    used to encapsulate the creation logic of an object or an aggregate
    """
    pass


class Repository():
    """
    used to fetch entities from the used data storage and save the information of entities to it.
    """
    pass



__all__ = ['TCPServiceClient', 'TCPApplicationService', 'TCPDomainService', 'TCPInfraService', 'HTTPServiceClient',
           'HTTPApplicationService', 'HTTPDomainService', 'HTTPInfraService', 'api', 'request', 'subscribe', 'publish',
           'get', 'post', 'head', 'put', 'patch', 'delete', 'options', 'trace', 'Entity', 'Value',
           'Aggregate', 'Factory', 'Repository', 'Bus', 'Registry', 'PostgresStore', 'cursor', 'dict_cursor',
           'nt_cursor', 'transaction',  'config_logs', 'setup_logging']

from .services import (TCPServiceClient, TCPApplicationService, TCPDomainService, TCPInfraService, HTTPServiceClient,
                       HTTPApplicationService, HTTPDomainService, HTTPInfraService, api, request, subscribe, publish,
                       get, post, head, put, patch, delete, options, trace)
from .model import (Entity, Value, Aggregate, Factory, Repository)
from .bus import Bus
from .registry import Registry
from .utils.log import setup_logging, config_logs
from .sql import PostgresStore, cursor, dict_cursor, nt_cursor, transaction



from uuid import uuid4

class _Packet:
    _pid = 0

    @classmethod
    def initialize(cls):
        import random
        cls._pid = random.randint(1000000, 9999999)

    @classmethod
    def _next_pid(cls):
        cls._pid += 1
        return cls._pid


class ControlPacket(_Packet):

    @classmethod
    def registration(cls, ip:str, port:int, node_id, service:str, version:str, vendors, service_type:str):
        v = []

        for vendor in vendors:
            v.append({'service': vendor.name, 'version': vendor.version})

        params = {'service': service,
                  'version': version,
                  'host': ip,
                  'port': port,
                  'node_id': node_id,
                  'vendors': v,
                  'type': service_type}

        packet = {'pid': cls._next_pid(), 'type': 'register', 'params': params}
        return packet

    @classmethod
    def instances(cls, service, version):
        params = {'service': service, 'version': version}

        packet = {'pid': cls._next_pid(),
                  'type': 'get_instances',
                  'service': service,
                  'version': version,
                  'params': params,
                  'request_id': str(uuid4())}

        return packet



import asyncio
from functools import partial
import json
import os
import logging
import signal

from again.utils import unique_hex
from retrial.retrial import retry
import aiohttp

from aiohttp.web import Application, Response

from .jsonprotocol import ServiceHostProtocol, ServiceClientProtocol
from .registryclient import RegistryClient
from .services import TCPServiceClient, HTTPServiceClient, HTTPApplicationService
from .pinger import Pinger
from .pubsub_handler import PubSubHandler
from .packet import ControlPacket
from .utils.log import setup_logging

HTTP = 'http'
TCP = 'tcp'

logger = logging.getLogger(__name__)


def _retry_for_client_conn(result):
    if isinstance(result, tuple):
        return not isinstance(result[0], asyncio.transports.Transport) or not isinstance(result[1],
                                                                                         asyncio.Protocol)
    return True


def _retry_for_pub(result):
    return not result


def _retry_for_exception(e):
    return True


class Bus:
    def __init__(self):

        self._registry_client = None

        self._client_protocols = {}
        self._pingers = {}
        self._service_clients = []
        self._node_clients = {}

        self._pending_requests = []
        self._unacked_publish = {}

        self._tcp_host = None
        self._http_host = None
        self._host_id = unique_hex()
        self._pubsub_handler = None
        self._ronin = False
        self._registered = False

    @property
    def ronin(self):
        return self._ronin

    @ronin.setter
    def ronin(self, value):
        self._ronin = value

    def require(self, args):
        for each in args:
            if isinstance(each, (TCPServiceClient, HTTPServiceClient)):
                each.bus = self
                self._service_clients.append(each)

    def serve_tcp(self, service_host):
        self._tcp_host = service_host
        self._tcp_host.bus = self

    def serve_http(self, service_host):
        self._http_host = service_host
        self._http_host.bus = self

    def send(self, packet:dict):
        packet['from'] = self._host_id
        func = getattr(self, '_' + packet['type'] + '_sender')
        func(packet)

    def send_http_request(self, app:str, service:str, version:str, method:str, entity:str, params:dict):
        """
        a convenience method that allows you to send a well formatted http request to another service
        """
        host, port, node_id, service_type = self._registry_client.resolve(service, version, entity, HTTP)

        url = 'http://{}:{}{}'.format(host, port, params.pop('path'))

        http_keys = ['data', 'headers', 'cookies', 'auth', 'allow_redirects', 'compress', 'chunked']
        kwargs = {k: params[k] for k in http_keys if k in params}

        query_params = params.pop('params', {})

        if app is not None:
            query_params['app'] = app

        query_params['version'] = version
        query_params['service'] = service

        response = yield from aiohttp.request(method, url, params=query_params, **kwargs)
        return response

    def _request_sender(self, packet: dict):
        """
        sends a request to a server from a ServiceClient
        auto dispatch method called from self.send()
        """
        self._pending_requests.append(packet)
        self._clear_request_queue()

    def host_receive(self, packet: dict, protocol: ServiceHostProtocol):
        if packet['type'] == 'ping':
            self._handle_ping(packet, protocol)
        elif packet['type'] == 'pong':
            self._handle_pong(packet['node_id'], packet['count'])
        elif packet['type'] == 'ack':
            pid = packet['pid']
            self._unacked_publish.pop(pid)
        else:
            if self._tcp_host.is_for_me(packet['service'], packet['version']):
                func = getattr(self, '_' + packet['type'] + '_receiver')
                func(packet, protocol)
            else:
                logger.warn('wrongly routed packet: ', packet)

    def _request_receiver(self, packet, protocol):
        api_fn = getattr(self._tcp_host, packet['endpoint'])
        if api_fn.is_api:
            from_node_id = packet['from']
            entity = packet['entity']
            future = asyncio.async(api_fn(from_id=from_node_id, entity=entity, **packet['payload']))

            def send_result(f):
                result_packet = f.result()
                protocol.send(result_packet)

            future.add_done_callback(send_result)
        else:
            print('no api found for packet: ', packet)

    def client_receive(self, service_client:TCPServiceClient, packet:dict):
        logger.info('service client {}, packet {}'.format(service_client, packet))
        logger.info('active pingers {}'.format(self._pingers))
        if packet['type'] == 'ping':
            self._handle_ping(packet['node_id'], packet['count'])
        elif packet['type'] == 'pong':
            pinger = self._pingers[packet['node_id']]
            asyncio.async(pinger.pong_received(packet['count']))
        else:
            service_client.process_packet(packet)

    def _stop(self, signame:str):
        print('\ngot signal {} - exiting'.format(signame))
        asyncio.get_event_loop().stop()

    def _host_factory(self):
        return ServiceHostProtocol(self)

    def _client_factory(self):
        return ServiceClientProtocol(self)

    def start(self, registry_host: str, registry_port: int, redis_host: str, redis_port: int):
        self._set_process_name()
        host = self._tcp_host if self._tcp_host else self._http_host
        setup_logging('{}_{}'.format(host.name, host.socket_address[1]))
        ControlPacket.initialize()
        asyncio.get_event_loop().add_signal_handler(getattr(signal, 'SIGINT'), partial(self._stop, 'SIGINT'))
        asyncio.get_event_loop().add_signal_handler(getattr(signal, 'SIGTERM'), partial(self._stop, 'SIGTERM'))

        tcp_server = self._create_tcp_service_host()
        http_server = self._create_http_service_host()
        self._create_pubsub_handler(redis_host, redis_port)
        if not self.ronin:
            self._setup_registry_client(registry_host, registry_port)
            if self._tcp_host:
                tcp_host_ip, tcp_host_port = self._tcp_host.socket_address
                self._registry_client.register_tcp(self._service_clients, tcp_host_ip, tcp_host_port,
                                               *self._tcp_host.properties)
            if self._http_host:
                ip, port = self._http_host.socket_address
                self._registry_client.register_http(self._service_clients, ip, port, *self._http_host.properties)

        if tcp_server:
            logger.info('Serving TCP on {}'.format(tcp_server.sockets[0].getsockname()))
        if http_server:
            logger.info('Serving HTTP on {}'.format(http_server.sockets[0].getsockname()))
        logger.info("Event loop running forever, press CTRL+c to interrupt.")
        logger.info("pid %s: send SIGINT or SIGTERM to exit." % os.getpid())

        try:
            asyncio.get_event_loop().run_forever()
        except Exception as e:
            print(e)
        finally:
            if tcp_server:
                tcp_server.close()
                asyncio.get_event_loop().run_until_complete(tcp_server.wait_closed())

            if http_server:
                http_server.close()
                asyncio.get_event_loop().run_until_complete(http_server.wait_closed())

            asyncio.get_event_loop().close()

    def _register_for_subscription(self):
        subscription_list = []
        for client in self._service_clients:
            if isinstance(client, TCPServiceClient):
                for each in dir(client):
                    fn = getattr(client, each)
                    if callable(fn) and getattr(fn, 'is_subscribe', False):
                        subscription_list.append((client.name, client.version, fn.__name__))
        yield from self._pubsub_handler.subscribe(subscription_list, handler=self.subscription_handler)

    def publish(self, service, version, endpoint, payload):
        asyncio.async(self._retry_publish(service, version, endpoint, payload))

    @retry(should_retry_for_result=_retry_for_pub, should_retry_for_exception=_retry_for_exception, timeout=10,
           strategy=[0, 2, 2, 4])
    def _retry_publish(self, service, version, endpoint, payload):
        return (yield from self._pubsub_handler.publish(service, version, endpoint, payload))

    def subscription_handler(self, service, version, endpoint, payload):
        client = [sc for sc in self._service_clients if (sc.name == service and sc.version == version)][0]
        func = getattr(client, endpoint)
        asyncio.async(func(**json.loads(payload)))

    def registration_complete(self):
        if not self._registered:
            f = self._create_service_clients()
            self._registered = True

            def fun(fut):
                if self._tcp_host:
                    self._clear_request_queue()

            f.add_done_callback(fun)

    def _create_tcp_service_host(self):
        if self._tcp_host:
            host_ip, host_port = self._tcp_host.socket_address
            host_coro = asyncio.get_event_loop().create_server(self._host_factory, host_ip, host_port)
            return asyncio.get_event_loop().run_until_complete(host_coro)

    def _create_http_service_host(self):
        if self._http_host:
            host_ip, host_port = self._http_host.socket_address
            ssl_context = self._http_host.ssl_context
            app = Application(loop=asyncio.get_event_loop())
            for each in self._http_host.__ordered__:
                fn = getattr(self._http_host, each)
                if callable(fn) and getattr(fn, 'is_http_method', False):
                    for path in fn.paths:
                        app.router.add_route(fn.method, path, fn)
                        if self._http_host.cross_domain_allowed:
                            app.router.add_route('options', path, self._http_host.preflight_response)
            fn = getattr(self._http_host, 'pong')
            app.router.add_route('GET', '/ping', fn)
            handler = app.make_handler()
            http_coro = asyncio.get_event_loop().create_server(handler, host_ip, host_port, ssl=ssl_context)
            return asyncio.get_event_loop().run_until_complete(http_coro)

    def _create_pubsub_handler(self, host, port):
        self._pubsub_handler = PubSubHandler(host, port)
        asyncio.get_event_loop().run_until_complete(self._pubsub_handler.connect())
        asyncio.async(self._register_for_subscription())

    def _create_service_clients(self):
        futures = []
        for sc in self._service_clients:
            for host, port, node_id, service_type in self._registry_client.get_all_addresses(sc.properties):
                self._node_clients[node_id] = sc
                future = self._connect_to_client(host, node_id, port, service_type)
                futures.append(future)
        return asyncio.gather(*futures, return_exceptions=False)

    @retry(should_retry_for_result=_retry_for_client_conn, should_retry_for_exception=_retry_for_exception, timeout=10,
           strategy=[0, 2, 2, 4])
    def _connect_to_client(self, host, node_id, port, service_type):
        logger.info('node_id' + node_id)
        future = asyncio.async(asyncio.get_event_loop().create_connection(self._client_factory, host, port))
        future.add_done_callback(
            partial(self._service_client_connection_callback, self._node_clients[node_id], node_id, service_type))
        return future

    def _service_client_connection_callback(self, sc, node_id, service_type, future):
        transport, protocol = future.result()
        protocol.set_service_client(sc)
        if service_type == TCP:
            pinger = Pinger(self, asyncio.get_event_loop())
            self._pingers[node_id] = pinger
            pinger.register_tcp_service(protocol, node_id)
            asyncio.async(pinger.start_ping())
        self._client_protocols[node_id] = protocol

    def _setup_registry_client(self, host: str, port: int):
        self._registry_client = RegistryClient(asyncio.get_event_loop(), host, port, self)
        self._registry_client.connect()

    @staticmethod
    def _create_json_service_name(app, service, version):
        return {'app': app, 'service': service, 'version': version}

    def _handle_ping(self, packet, protocol):
        pong_packet = self._make_pong_packet(packet['node_id'], packet['count'])
        protocol.send(pong_packet)

    def _handle_pong(self, node_id, count):
        pinger = self._pingers[node_id]
        asyncio.async(pinger.pong_received(count))

    def _make_pong_packet(self, node_id, count):
        packet = {'type': 'pong', 'node_id': node_id, 'count': count}
        return packet

    def _clear_request_queue(self):
        self._pending_requests[:] = [each for each in self._pending_requests if not self._send_packet(each)]

    def _send_packet(self, packet):
        node_id = self._get_node_id_for_packet(packet)
        if node_id is not None:
            client_protocol = self._client_protocols[node_id]
            if client_protocol.is_connected:
                packet['to'] = node_id
                client_protocol.send(packet)
                return True
            else:
                return False
        return False

    def _get_node_id_for_packet(self, packet):
        app, service, version, entity = packet['app'], packet['service'], packet['version'], packet['entity']
        node = self._registry_client.resolve(service, version, entity, TCP)
        return node[2] if node else None

    @staticmethod
    def send_ack(protocol, pid):
        packet = {'type': 'ack', 'pid': pid}
        protocol.send(packet)

    def handle_ping_timeout(self, node_id):
        logger.info("Service client connection timed out {}".format(node_id))
        self._pingers.pop(node_id, None)
        service_props = self._registry_client.get_for_node(node_id)
        logger.info('service client props {}'.format(service_props))
        if service_props is not None:
            host, port, _node_id, _type = service_props
            asyncio.async(self._connect_to_client(host, _node_id, port, _type))

    def _set_process_name(self):
        from setproctitle import setproctitle

        if self._tcp_host:
            setproctitle('{}_{}_{}'.format(self._tcp_host.name, self._tcp_host.version, self._host_id))
        elif self._http_host:
            setproctitle('{}_{}_{}'.format(self._http_host.name, self._http_host.version, self._host_id))


if __name__ == '__main__':
    REGISTRY_HOST = '127.0.0.1'
    REGISTRY_PORT = 4500
    REDIS_HOST = '127.0.0.1'
    REDIS_PORT = 6379
    HOST_IP = '127.0.0.1'
    HOST_PORT = 8000
    bus = Bus()
    bus.start(REGISTRY_HOST, REGISTRY_PORT, REDIS_HOST, REDIS_PORT)
